#include "stdafx.h"
#include <cstdarg>
#include <ctype.h>
#include "assembler.h"

#pragma warning(disable: 4996)

bool Assembler::Assemble(char* inputFilename, FILE* pInput, uint8vec* outputBuffer)
{
	m_inputFilename = inputFilename;
	m_outputBuffer = outputBuffer;

	for (m_lineNumber = 1; fgets(m_inputLine, kInputLineLength-1, pInput); m_lineNumber++)
	{
		// initialize tokenization
		m_inputLineLength = strlen(m_inputLine);
		m_tokenEnd = -1;

		while (1)
		{
			// end of line?
			char* first = ReadToken();
			int firstLen = strlen(first);
			if (firstLen == 0)
				break;

			// OK to assemble program binaries larger than 1K in size
                     //if (GetCurrentOutputAddress() >= 1<<(16-kOpcodeSize))
			//{
			//	ReportError("assembly address out of range");
			//	return false;
			//}

			// is it a mnemonic?
			if (OpcodeEntry* pEntry = GetOpcodeTableEntry(first))
			{
				char* mnemo = first;
				char* operand = ReadToken();

				if (operand[0] == '\0')
				{
					if (!ParseImplied(pEntry, mnemo))
						return false;
				}
				else if (operand[0] == '#')
				{
					if (!ParseImmediate(pEntry, mnemo, operand))
						return false;
				}
				else
				{
					if (!ParseAddressed(pEntry, mnemo, operand))
						return false;
				}
			}
			// is it a label?
			else if (first[0] == '+' || first[0] == '-' || first[firstLen-1] == ':')
			{
				if (!ParseLabel(first))
					return false;
			}
                     // is it a define?
                     else if (stricmp(first, "#define") == 0)
                     {
                        char* name = ReadToken();
                        if (name[0] == '\0')
                        {
                            ReportError("missing #define constant name");
                            return false;
                        }
                        
                        char* value = ReadToken();
                        if (!ParseDefine(name, value))
                            return false;
                     }
                     // is it a directive?
                     else if (first[0] == '.')
                     {
                        if (!ParseDirective(first))
                            return false;
                     }
			else
			{
				ReportError("unrecognized identifier '%s'", first);
				return false;
			}
		}
	}

	// report any unresolved references
	if (!m_pendingReferences.empty())
	{
		for (PendingReferenceList::iterator it = m_pendingReferences.begin(); it != m_pendingReferences.end(); ++it)
		{
			m_lineNumber = it->second[0].lineNumber; // report the line number of the first reference to this label
			ReportError("could not resolve label '%s'", it->first.c_str());
		}

		return false;
	}

	return true;
}

bool Assembler::ParseImplied(OpcodeEntry* pEntry, char* mnemo)
{
	uint8 opcode = pEntry->opcodes[kImplied];
	if (opcode != kInvalidOp)
	{
		m_outputBuffer->push_back(opcode << (8-kOpcodeSize));
		//m_outputBuffer->push_back(0);
	}
	else
	{
		ReportError("implied addressing not supported for '%s'", mnemo);
		return false;
	}

	return true;
}

bool Assembler::ParseImmediate(OpcodeEntry* pEntry, char* mnemo, char* operand)
{
	uint8 opcode = pEntry->opcodes[kImmediate];
	if (opcode != kInvalidOp)
	{
		uint8 immValue;
		if (operand[1] == '$')
			immValue = (uint8) strtoul(&operand[2], NULL, 16);
		else if (isdigit(operand[1]))
			immValue = (uint8) strtoul(&operand[1], NULL, 10);
              else
              {
                string constantName(&operand[1]);
                LabelToAddressMap::iterator labelIt = m_labels.find(constantName);
                if (labelIt == m_labels.end())
                {
                    ReportError("undefined constant '%s'", &operand[1]);
                    return false;
                }
                int constValue = m_labels[constantName];
                if (constValue > 255 || constValue < -128)
                {
                    ReportError("constant '%s' value exceeds size of byte", &operand[1]);
                    return false;
                }
                if (constValue < 0)
                {
                    constValue += 256;
                }
                immValue = constValue;
              }

		m_outputBuffer->push_back(opcode << (8-kOpcodeSize));
		m_outputBuffer->push_back(immValue);
	}
	else
	{
		ReportError("immediate addressing not supported for '%s'", mnemo);
		return false;
	}

	return true;
}

bool Assembler::ParseAddressed(OpcodeEntry* pEntry, char* mnemo, char* operand)
{
	int address;
	if (operand[0] == '$')
	{
		address = (int) strtoul(&operand[1], NULL, 16);
	}
	else if (isdigit(operand[0]))
	{
		address = (int) strtoul(&operand[0], NULL, 10);
	}
	else 
	{
		LabelToAddressMap::iterator labelIt = m_labels.find(operand);
		if (labelIt != m_labels.end())
		{
			// existing label match
			address = labelIt->second;
		}
		else
		{
			// forward label reference
			m_pendingReferences[operand].push_back(PendingReference(m_lineNumber, GetCurrentOutputAddress()));
			address = 0;
		}
	}

	if (address >= 1<<(16-kOpcodeSize))
	{
		ReportError("operand address out of range");
		return false;
	}

	uint8 opcode;
	int operandLen = strlen(operand);
	if (operandLen >= 3 && stricmp(&operand[operandLen-2], ",x") == 0)
	{
		// indexed
		opcode = pEntry->opcodes[kIndexed];
		if (opcode == kInvalidOp)
		{
			ReportError("indexed addressing not supported for '%s'", mnemo);
			return false;
		}
	}
	else
	{
		// absolute
		opcode = pEntry->opcodes[kAbsolute];
		if (opcode == kInvalidOp)
		{
			ReportError("absolute addressing not supported for '%s'", mnemo);
			return false;
		}
	}

	m_outputBuffer->push_back((opcode << (8-kOpcodeSize)) | (address >> 8));
	m_outputBuffer->push_back(address & 0xff);

	return true;
}

bool Assembler::ParseLabel(char* label)
{
	// discard the trailing colon for named labels
	if (label[0] != '+' && label[0] != '-')
	{
		label[strlen(label)-1] = '\0';
	}
	
	string labelName(label);
	int labelAddress = GetCurrentOutputAddress() % kBlockSize;

	// if it's not a +ahead label, then save it
	if (label[0] != '+')
	{
		LabelToAddressMap::iterator labelIt = m_labels.find(labelName);
		if (labelIt != m_labels.end())
		{
			// if it's not a -back label, report an error
			if (label[0] != '-')
			{
				ReportError("label '%s' was already defined in this context", label);
				return false;
			}
		}

		m_labels[label] = labelAddress;
	}

	// if it's not a -back label, resolve any pending references to this label
	if (label[0] != '-')
	{
		PendingReferenceList::iterator referenceIt = m_pendingReferences.find(labelName);
		if (referenceIt != m_pendingReferences.end())
		{
			vector<PendingReference>& referenceAddresses = referenceIt->second;
			for (vector<PendingReference>::iterator it = referenceAddresses.begin(); it != referenceAddresses.end(); ++it)
			{
				int addressToFix = it->outputAddress;
				(*m_outputBuffer)[addressToFix] |= (labelAddress >> 8);
				(*m_outputBuffer)[addressToFix+1] = labelAddress & 0xff;
			}

			m_pendingReferences.erase(referenceIt);
		}
	}

	return true;
}

bool Assembler::ParseDefine(char* name, char* valueString)
{
    if (valueString[0] == '\0')
    {
        ReportError("missing #define value");
        return false;
    }

    string constantName(name);
    LabelToAddressMap::iterator labelIt = m_labels.find(constantName);
    if (labelIt != m_labels.end())
    {
        ReportError("constant '%s' was already defined in this context", name);
        return false;
    }

    int value;
    if (valueString[0] == '$')
    {
        value = (int) strtoul(&valueString[1], NULL, 16);
    }
    else if (isdigit(valueString[0]))
    {
        value = (int) strtoul(&valueString[0], NULL, 10);
    }
    else
    {
        ReportError("constant definition by another constant is unimplemnted");
        return false;
    }

    m_labels[name] = value;

    return true;
}


bool Assembler::ParseDirective(char* directive)
{
    if (stricmp(directive, ".org") == 0)
    {
        char* valueString = ReadToken();
        if (valueString[0] == '\0')
        {
            ReportError("missing address");
            return false;
        }

        int value;
        if (valueString[0] == '$')
        {
            value = (int) strtoul(&valueString[1], NULL, 16);
        }
        else if (isdigit(valueString[0]))
        {
            value = (int) strtoul(&valueString[0], NULL, 10);
        }
        else
        {
            ReportError(".org definition by a constant is unimplemnted");
            return false;
        }

        int currentAddress = GetCurrentOutputAddress();
        if (value < currentAddress)
        {
            ReportError(".org address %X is less than current address %X", value, currentAddress);
            return false;
        }

        for (int i=0; i<value-currentAddress; i++)
        {
            m_outputBuffer->push_back(0);
        }
    }
    else
    {
        ReportError("unrecognized directive '%s'", directive);
        return false;
    }

    return true;
}

void Assembler::ReportError(char* format, ...)
{
	va_list args;
    va_start(args, format);
	printf("%s(%d) : ", m_inputFilename, m_lineNumber);
	vprintf(format, args);
	printf("\n");
	va_end(args);
}

Assembler::OpcodeEntry* Assembler::GetOpcodeTableEntry(const char* mnemo) const
{
	static OpcodeEntry opcodeTable[] = 
	{
		//  op   impl   abs   imm  indx
		{ "sub", 0xff, 0x00, 0x01, 0x02 },
		{ "add", 0xff, 0x04, 0x05, 0x06  },
		{ "cmp", 0xff, 0x08, 0x09, 0x0a  },
		{ "nor", 0xff, 0x0c, 0x0d, 0x0e  },
		{ "lda", 0xff, 0x10, 0x11, 0x12  },
		{ "sta", 0xff, 0x14, 0x15, 0x16  },
		{ "ldx", 0xff, 0x18, 0x19, 0x1a  },
		{ "stx", 0xff, 0x1c, 0x1d, 0x1e  },
		{ "cpx", 0xff, 0x1a, 0x1b, 0xff  },
		{ "bcc", 0xff, 0x2a, 0xff, 0xff  },
		{ "bcs", 0xff, 0x2e, 0xff, 0xff  },
		{ "bne", 0xff, 0x22, 0xff, 0xff  },
		{ "beq", 0xff, 0x26, 0xff, 0xff  },
		{ "pla", 0x30, 0xff, 0xff, 0xff  },
		{ "plx", 0x31, 0xff, 0xff, 0xff  },
		{ "return", 0x32, 0xff, 0xff, 0xff  },
		{ "pha", 0x34, 0xff, 0xff, 0xff  },
		{ "phx", 0x35, 0xff, 0xff, 0xff  },
		{ "jmp", 0xff, 0x36, 0xff, 0xff  },
		{ "call", 0xff, 0x37, 0xff, 0xff  },
		{ "inx", 0x38, 0xff, 0xff, 0xff  },
		{ "dex", 0x39, 0xff, 0xff, 0xff  },
		{ "halt", 0x3f, 0xff, 0xff, 0xff  },
	};

	for (int i=0; i < sizeof(opcodeTable)/sizeof(OpcodeEntry); i++)
	{
		if (stricmp(mnemo, opcodeTable[i].name) == 0)
			return &opcodeTable[i];
	}

	return NULL;
}

char* Assembler::ReadToken()
{
	// already at end?
	if (m_tokenEnd+1 >= m_inputLineLength)
	{
		return &m_inputLine[m_inputLineLength];
	}

	// skip whitespace
	int nextTokenStart = m_tokenEnd + 1;
	while (isspace(m_inputLine[nextTokenStart]))
	{
		nextTokenStart++;
	}

	// anything left?
	if (nextTokenStart != m_inputLineLength)
	{
		// proceed to whitespace, comment, or EOL
		int nextTokenEnd = nextTokenStart+1;
		while (m_inputLine[nextTokenEnd] != '\0' && !isspace(m_inputLine[nextTokenEnd]) && m_inputLine[nextTokenEnd] != ';')
		{
			nextTokenEnd++;
		}

		// mark end of token
		m_tokenEnd = nextTokenEnd;
		m_inputLine[m_tokenEnd] = '\0';
	}

	// if token is a comment, skip to EOL
	if (m_inputLine[nextTokenStart] == ';')
	{
		nextTokenStart = m_inputLineLength;
		m_tokenEnd = m_inputLineLength;
	}

	return &m_inputLine[nextTokenStart];
}