`include "defs.v"

module tinycpu
	#(parameter addrSize = 10)
	(input clk,
	 input reset,
	 inout [7:0] dataBus,
	 output reg [addrSize-1:0] addrBus,
	 output weMem);

	// CPU registers
	reg [7:0] aReg;
	reg [7:0] xReg;
	reg [5:0] spReg;
	reg [addrSize-1:0] pcReg;
	reg [addrSize-1:0] eaReg;
	reg cFlagReg, zFlagReg;
	reg [5:0] opcode;
	reg [3:0] state;
			
	// control signals
	reg [3:0] nextState;
	reg [7:0] dataBusOutput;
	wire [7:0] addSubResult;
	wire addSubCarryOut;
		
	// convenience methods
	wire [2:0] mathFunc = opcode[4:2];
	wire isMathOp = !opcode[5];
	wire isIndexed = opcode[1];
	wire isImmediate = opcode[0];
	wire isBranch = (opcode[5:4] == 2'b10);
	wire isCpx = opcode[4:1] == 4'b1101;
	wire runAlu = (state == `sFetchInstruction1 && isMathOp && isImmediate) || (state == `sDoMath); 
	wire flagsMatch = (opcode[3] && opcode[2] && cFlagReg) || (!opcode[3] && opcode[2] && zFlagReg) ||
				      (opcode[3] && !opcode[2] && !cFlagReg) || (!opcode[3] && !opcode[2] && !zFlagReg);
						 	               
	// -- current state
	always @(posedge clk or negedge reset) begin
		if (!reset)
			state <= `sFetchInstruction0;
		else 
			state <= nextState; 
	end

	// -- opcode
	always @(posedge clk) begin
		if (state == `sFetchInstruction0)
			opcode <= dataBus[7:2]; 
	end
	
	// -- stack pointer
	// can increment or decrement
	always @(posedge clk or negedge reset) begin
		if (!reset)
			spReg <= 6'b111111;
		else if (state == `sReturn0 ||
		         (state == `sFetchInstruction0 && dataBus[7:4] == 4'b1100)) // pla, plx, return
			spReg <= spReg + 1'b1;
		else if (state == `sPushA ||
		         state == `sPushX ||
		         state == `sCall0 ||
		         state == `sCall1)
			spReg <= spReg - 1'b1;		
	end
	
	// -- X (index) register
	// can increment, decrement, or load	
	always @(posedge clk) begin
		if (state == `sFetchInstruction0 && dataBus[7:2] == 6'b111000) // inx
			xReg <= xReg + 1'b1;
		else if (state == `sFetchInstruction0 && dataBus[7:2] == 6'b111001) // dex
			xReg <= xReg - 1'b1;
		else if ((runAlu && opcode[4:1] == 4'b1100) || // ldx or plx
		         (state == `sPullX))
			xReg <= dataBus;						
	end
	
	// -- program counter
	// can increment or load
	always @(posedge clk or negedge reset) begin
		if (!reset)
			pcReg <= 0; 
		else if (state == `sFetchInstruction0 ||
	             (state == `sFetchInstruction1 && opcode != 6'b110110 && !isBranch) || // not jump or branch
	             (state == `sFetchInstruction1 && isBranch && !flagsMatch)) // branch not taken
			pcReg <= pcReg + 1'b1;
		else if ((state == `sReturn1) ||
	             (state == `sFetchInstruction1 && opcode == 6'b110110) || // jump
	             (state == `sFetchInstruction1 && isBranch && flagsMatch)) // branch taken
			pcReg <= { eaReg[addrSize-1:8], dataBus };
		else if (state == `sCall1)
			pcReg <= eaReg;
	end

	// -- effective address register
	always @(posedge clk) begin
		if (state == `sAddIndex)
			eaReg <= eaReg + xReg;	
		else if (state == `sFetchInstruction0 || state == `sReturn0)
			eaReg[addrSize-1:8] <= dataBus[addrSize-9:0];
		else if (state == `sFetchInstruction1)
			eaReg[7:0] <= dataBus;
	end
	
	// -- Adder/subtractor
	wire [7:0] aluLeft = isCpx ? xReg : aReg;
	assign { addSubCarryOut, addSubResult } = 
		mathFunc[0] ? 
			aluLeft + dataBus[7:0] : 
			aluLeft + {1'b0, ~dataBus[7:0]} + 1'b1; // add two's complement
	
	// -- A (accumulator) register
	// can add, subtract, NOR, or load	
	always @(posedge clk) begin
		if (runAlu && !mathFunc[2] && !mathFunc[1]) // add or sub
			aReg <= addSubResult;
		else if (runAlu && mathFunc == 3'b011) // nor
			aReg <= ~(aReg | dataBus); 
		else if ((runAlu && mathFunc == 3'b100) || // lda or pla
	             (state == `sPullA))
			aReg <= dataBus;		
	end
	
	// -- flags registers
	always @(posedge clk) begin
		if ((runAlu && !mathFunc[2] && !mathFunc[1]) ||
		    (runAlu && mathFunc == 3'b010) ||
		    (runAlu && opcode[4:1] == 4'b1101)) begin // performing add, sub, cmp, or cpx
			cFlagReg <= addSubCarryOut;
			zFlagReg <= (addSubResult == 0); 
		end 
		else if (state == `sFetchInstruction0 && isMathOp && (mathFunc == 3'b011 || mathFunc == 3'b100)) // just finished nor or lda
			zFlagReg <= (aReg == 0); 
		else if (state == `sFetchInstruction0 && 
		         (opcode[5:1] == 5'b11100 || opcode[5:1] == 5'b01100)) // just finished inx, dex, or ldx
			zFlagReg <= (xReg == 0); 
	end
					
	// -- address output
	always @* begin
		case (state)
			`sDoMath: 
				addrBus <= eaReg;
			`sPushA, `sPushX, `sPullA, `sPullX, `sCall0, `sCall1, `sReturn0, `sReturn1: 
				addrBus <= { {(addrSize-6){1'b1}}, spReg }; // pad address with leading 1's
			`sFetchInstruction0, `sFetchInstruction1: 
				addrBus <= pcReg;
			default: 
				addrBus <= {{addrSize}{1'bx}};
		endcase
	end
	 	
	// -- data output
	always @* begin
		case (state)
			`sDoMath:
				dataBusOutput <= mathFunc[1] ? xReg : aReg;
			`sPushA:
				dataBusOutput <= aReg;
			`sPushX:
				dataBusOutput <= xReg;
			`sCall0:
				dataBusOutput <= pcReg[7:0];
			`sCall1:
				dataBusOutput <= { {(16-addrSize){1'b0}}, pcReg[addrSize-1:8] };
			default:
				dataBusOutput <= 8'bxxxxxxxx;
		endcase
	end
	
	wire writeCycle =  (state == `sPushA ||
						state == `sPushX || 
						state == `sCall0 || 
						state == `sCall1 ||
						state == `sDoMath && mathFunc[2] && mathFunc[0]); // sta or stx
	                
	assign weMem = !(writeCycle & !clk); // only assert we during 2nd half of the clock cycle
	assign dataBus = !weMem ? dataBusOutput : 8'bzzzzzzzz;
	
	// -- control logic: state machine
	always @* begin		
		case (state)
			`sFetchInstruction0:			
				// could save a clock cycle (and a byte) for pha, phx, pla, plx, inx, dex, return, by skipping sFetchInstruction1 here, but
				// complicates state logic and increases resource needs
				nextState <= (dataBus[7:3] == 5'b11100) ? `sFetchInstruction0 : // inx, dex
							 (dataBus[7:2] == 6'b110010) ? `sReturn0 : // return
							 (dataBus[7:2] == 6'b110100) ? `sPushA : // pha
							 (dataBus[7:2] == 6'b110101) ? `sPushX : // phx
							 (dataBus[7:2] == 6'b110000) ? `sPullA : // pla
							 (dataBus[7:2] == 6'b110001) ? `sPullX : // plx				
							 `sFetchInstruction1;
			`sFetchInstruction1: 
				nextState <= (isMathOp & !isImmediate & isIndexed & !isCpx) ? `sAddIndex : // math op, indexed (cpx abs encodes as ldx abs,x)
							 (isMathOp & !isImmediate & (!isIndexed | isCpx)) ? `sDoMath : // math op, absolute (cpx abs encodes as ldx abs,x)
							 (isMathOp & isImmediate) ? `sFetchInstruction0 : // math op, immediate mode
							 (opcode == 6'b110110) ? `sFetchInstruction0 : // jump
							 isBranch ? `sFetchInstruction0 : // branch			 
							 (opcode == 6'b110111) ? `sCall0 : // call
							 4'bxxxx; // invalid opcode
			`sAddIndex: 
				// ea <= ea+x
				nextState <= `sDoMath;		
			`sDoMath: 
				// load/store/ALU					 			
				nextState <= `sFetchInstruction0;			
			`sPushA:
				// addrsel=sp, dbus <= a, sp <= sp-1 
				nextState <= `sFetchInstruction0;	
			`sPushX:
				// addrsel=sp, dbus <= x, sp <= sp-1 
				nextState <= `sFetchInstruction0;	
			`sPullA:
				// addrsel=sp, a <= dbus
				nextState <= `sFetchInstruction0;
			`sPullX:
				// addrsel=sp, x <= dbus
				nextState <= `sFetchInstruction0;
			`sCall0:
				// addrsel=sp, dbus <= pclo, sp <= sp-1 
				nextState <= `sCall1;
			`sCall1:
				// addrsel=sp, dbus <= pchi, sp <= sp-1
				nextState <= `sFetchInstruction0;		
			`sReturn0:
				// addrsel=sp, ea <= dbus, sp <= sp+1
				nextState <= `sReturn1;
			`sReturn1:
				// addrsel=sp, pc <= {ea,dbus}
				nextState <= `sFetchInstruction0;
		endcase
	end	
	 
endmodule